#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <Rdefines.h>

#include "calcalcs.h"
#include "utCalendar2_cal.h"
#include "udunits.h"

void calcalcs_cal_finalizer( SEXP ptr );
void R_ccs_error_message( int errnum ) ;

/*************************************************************************************************
 * Returns an integer that is 0 on success, and -1 on error
 * An attribute of the returned vlaue holds the pointer to the calendar (if successful)
 */
SEXP R_ccs_init_calendar( SEXP sx_calname ) 
{
	calcalcs_cal	*retval;
	SEXP		sx_retval, retptr;

	const char *calname = CHAR(STRING_ELT(sx_calname,0));

	retval = ccs_init_calendar( calname );

	PROTECT( sx_retval = allocVector(INTSXP, 1));
	PROTECT( retptr = R_MakeExternalPtr( retval, install("calcalcs_cal_ptr"), R_NilValue ));
	R_RegisterCFinalizerEx( retptr, calcalcs_cal_finalizer, TRUE );

	if( retval == NULL ) 
		INTEGER(sx_retval)[0] = -1;	/* Error */
	else
		INTEGER(sx_retval)[0] = 0;	/* Success */

	setAttrib( sx_retval, install("calcalcs_cal_pointer"), retptr );
	UNPROTECT( 2 );

	return( sx_retval );
}

/*************************************************************************************************/
void calcalcs_cal_finalizer( SEXP ptr ) 
{
	calcalcs_cal	*pp;

	if( ! R_ExternalPtrAddr(ptr))
		return;

	pp = R_ExternalPtrAddr( ptr );
	ccs_free_calendar( pp );
	R_ClearExternalPtr(ptr);
}

/*************************************************************************************************
 * Returns a two-element integer array. First element is 0 on success, -1 on error. Second
 * is 1 if the year is a leap year in the specified calendar, and 0 otherwise.
 */
SEXP R_ccs_isleap( SEXP sx_calcalcs_cal_ptr, SEXP sx_year )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, isleap, ierr;

	PROTECT( sx_retval = allocVector(INTSXP, 2));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_isleap: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_isleap: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year = INTEGER(sx_year)[0];
	if( (ierr = ccs_isleap( calendar, year, &isleap )) != 0) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_isleap: error returned from call to calencals library routine ccs_isleap: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = isleap;
	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a two-element integer array. First element is 0 on success, -1 on error. Second
 * is the (true) Julian day number (count starting at 1 = 1 Jan 4713 BC).
 */
SEXP R_ccs_date2jday( SEXP sx_calcalcs_cal_ptr, SEXP sx_year, SEXP sx_month, SEXP sx_day )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, month, day, ierr, jday;

	PROTECT( sx_retval = allocVector(INTSXP, 2));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year  = INTEGER(sx_year )[0];
	month = INTEGER(sx_month)[0];
	day   = INTEGER(sx_day  )[0];

	ierr = ccs_date2jday( calendar, year, month, day, &jday );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: error returned from call to calencals library routine ccs_date2jday: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = jday;
	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a four-element integer array. First element is 0 on success, -1 on error. Second
 * is the year, third is the month, fourth is the day
 */
SEXP R_ccs_jday2date( SEXP sx_calcalcs_cal_ptr, SEXP sx_jday )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, month, day, ierr, jday;

	PROTECT( sx_retval = allocVector(INTSXP, 4));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	jday = INTEGER(sx_jday  )[0];

	ierr = ccs_jday2date( calendar, jday, &year, &month, &day );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_jday2date: error returned from call to calencals library routine ccs_jday2date: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = year;
	INTEGER(sx_retval)[2] = month;
	INTEGER(sx_retval)[3] = day;

	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a two-element integer array. First element is 0 on success, -1 on error. Second
 * is the number of days in the specified month
 */
SEXP R_ccs_dpm( SEXP sx_calcalcs_cal_ptr, SEXP sx_year, SEXP sx_month )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, month, ierr, dpm;

	PROTECT( sx_retval = allocVector(INTSXP, 2));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dpm: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dpm: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year  = INTEGER(sx_year )[0];
	month = INTEGER(sx_month)[0];

	ierr = ccs_dpm( calendar, year, month, &dpm );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dpm: error returned from call to calencals library routine ccs_dpm: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = dpm;
	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a two-element integer array. First element is 0 on success, -1 on error. Second
 * is the day number of the year, with counting starting at 1
 */
SEXP R_ccs_date2doy( SEXP sx_calcalcs_cal_ptr, SEXP sx_year, SEXP sx_month, SEXP sx_day )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, month, day, ierr, doy;

	PROTECT( sx_retval = allocVector(INTSXP, 2));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2doy: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2doy: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year  = INTEGER(sx_year )[0];
	month = INTEGER(sx_month)[0];
	day   = INTEGER(sx_day  )[0];

	ierr = ccs_date2doy( calendar, year, month, day, &doy );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_date2jday: error returned from call to calencals library routine ccs_date2doy: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = doy;
	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a four-element integer array. First element is 0 on success, -1 on error. Second
 * is the year, third is month, fourth is day
 */
SEXP R_ccs_doy2date( SEXP sx_calcalcs_cal_ptr, SEXP sx_year, SEXP sx_doy )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar;
	int		year, month, day, ierr, doy;

	PROTECT( sx_retval = allocVector(INTSXP, 4));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_doy2date: not passed a valid, initialized calendar\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_ptr, install("calcalcs_cal_pointer" )));
	if( calendar == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_doy2date: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year = INTEGER( sx_year )[0];
	doy  = INTEGER( sx_doy  )[0];

	ierr = ccs_doy2date( calendar, year, doy, &month, &day );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_doy2date: error returned from call to calencals library routine ccs_doy2date: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = year;
	INTEGER(sx_retval)[2] = month;
	INTEGER(sx_retval)[3] = day;

	UNPROTECT(1);
	return( sx_retval );
}

/*************************************************************************************************
 * Returns a four-element integer array. First element is 0 on success, -1 on error. Second,
 * third, and fourth elements are returned year, month, and day.
 */
SEXP R_ccs_dayssince( SEXP sx_calcalcs_cal_orig_ptr, SEXP sx_year, SEXP sx_month, SEXP sx_day, 
			SEXP sx_ndays_since, SEXP sx_calcalcs_cal_new_ptr )
{
	SEXP		sx_retval;
	calcalcs_cal	*calendar_orig, *calendar_new;
	int		year, month, day, ierr, ndays_since;
	int		year_new, month_new, day_new;

	PROTECT( sx_retval = allocVector(INTSXP, 4));
	INTEGER(sx_retval)[0] = 0;	

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_orig_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dayssince: not passed a valid, initialized calendar_orig\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar_orig = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_orig_ptr, install("calcalcs_cal_pointer" )));
	if( calendar_orig == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dayssince: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	if( ! R_ExternalPtrAddr(sx_calcalcs_cal_new_ptr)) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dayssince: not passed a valid, initialized calendar_new\n" );
		UNPROTECT(1);
		return (sx_retval );
		}

	calendar_new = R_ExternalPtrAddr( getAttrib(sx_calcalcs_cal_new_ptr, install("calcalcs_cal_pointer" )));
	if( calendar_new == NULL ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dayssince: not passed a valid, initialized calendar: %s\n", 
			ccs_err_str(ierr) );
		UNPROTECT(1);
		return (sx_retval );
		}

	year  = INTEGER(sx_year )[0];
	month = INTEGER(sx_month)[0];
	day   = INTEGER(sx_day  )[0];
	ndays_since = INTEGER( sx_ndays_since )[0];

	ierr = ccs_dayssince( calendar_orig, year, month, day, ndays_since, 
		calendar_new, &year_new, &month_new, &day_new );
	if( ierr != 0 ) {
		INTEGER(sx_retval)[0] = -1;	/* Error */
		Rprintf( "R_ccs_dayssince: error returned from call to calencals library routine ccs_dayssince: %s\n", 
			ccs_err_str(ierr) );
		/* R_ccs_error_message( ierr ); */
		UNPROTECT(1);
		return (sx_retval );
		}

	INTEGER(sx_retval)[1] = year_new;
	INTEGER(sx_retval)[2] = month_new;
	INTEGER(sx_retval)[3] = day_new;

	UNPROTECT(1);
	return( sx_retval );
}


/*************************************************************************************************/
void R_ccs_error_message( int errnum ) 
{
	if( errnum == 0 ) 
		return;

	if( errnum == CALCALCS_ERR_NO_YEAR_ZERO )
		Rprintf( "specified calendar has no valid year 0\n" );

	else if( errnum == CALCALCS_ERR_DATE_NOT_IN_CALENDAR )
		Rprintf( "passed date is not a valid date in the specified calendar\n" );

	else if( errnum == CALCALCS_ERR_INVALID_DAY_OF_YEAR )
		Rprintf( "passed day of year is not a valid day of year in the specified calendar\n" );

	else if( errnum == CALCALCS_ERR_NOT_A_MIXED_CALENDAR )
		Rprintf( "passed calendar is not a mixed calendar (e.g., one that transitions from Julian to Gregorian calendar\n");

	else if( errnum == CALCALCS_ERR_UNKNOWN_COUNTRY_CODE )
		Rprintf( "was given an unknown country code -- I do not have the calendar transition date for that code\n" );

	else if( errnum == CALCALCS_ERR_OUT_OF_RANGE )
		Rprintf( "value out of range\n" );

	else if( errnum == CALCALCS_ERR_NULL_CALENDAR )
		Rprintf( "was passed a NULL calendar!\n" );

	else if( errnum == CALCALCS_ERR_INVALID_CALENDAR )
		Rprintf( "was passed an invalid calendar!\n" );
}

/***************************************************************************************************/
/* Inputs:
 *	sx_value: guaranteed to be a double (could be >1)
 *		calendar date.
 *	sx_unit_str : unit string (Ex: "days since 1900-01-01" or "hours since 2000-01-01 00:00 -8")
 * 	calendar: either 'standard', 'noleap', or '360_day'
 *
 * Return value:
 * 	If only 1 value is passed, then returns an object of class "utDate"
 *	If more than 1 value passed, the returned value depends on the 'style' arg.
 *		style=1: returns a list of n objects of class "utDate"
 *		style=2: returns a list with year, month, day, hour, minute, second, each
 *			of which is an array of length n,
 *		where n is the number of values passed in.
 */
SEXP R_utCalendar_ccs_v3p0( SEXP sx_value, SEXP sx_unit_str, SEXP sx_style,
	SEXP sx_calendar )
{
	ut_unit *u;
	int 	i, nvals, retval, year, month, day, hour, minute, *style;
	double	second;
	double	*value, d_tzoffset_hours;
	float	f_tzoffset_hours;
	SEXP	sx_big_retval, sx_single_retval, sx_year, sx_month, sx_day, sx_hour, sx_minute, 
		sx_name, sx_second, sx_tzoffset_hours, sx_retarr_year, sx_retarr_month, sx_retarr_day, sx_retarr_hour,
		sx_retarr_minute, sx_retarr_second, sx_retarr_tzoffset_hours, sx_calendar_elt;
	const char *calendar = CHAR(STRING_ELT(sx_calendar,0));
	const char *unit_str = CHAR(STRING_ELT(sx_unit_str,0));

	nvals = length( sx_value );
	if( nvals < 1 ) 
		error( "utCalendar_v2p0 (R version): error: passed funny # of vals to convert: %d\n", nvals );

	value            = REAL(sx_value);
	style            = INTEGER(sx_style);

	/* If we are doing multiple entries, make our return object(s) */
	if( nvals > 1 ) {

		if( *style == 1 ) 
			PROTECT( sx_big_retval = allocVector( VECSXP, nvals )); /* a list that will hold utDate objects */

		else if(*style == 2 ) {
			PROTECT( sx_retarr_year   = allocVector( INTSXP,  nvals ));
			PROTECT( sx_retarr_month  = allocVector( INTSXP,  nvals ));
			PROTECT( sx_retarr_day    = allocVector( INTSXP,  nvals ));
			PROTECT( sx_retarr_hour   = allocVector( INTSXP,  nvals ));
			PROTECT( sx_retarr_minute = allocVector( INTSXP,  nvals ));
			PROTECT( sx_retarr_second = allocVector( REALSXP, nvals ));
			PROTECT( sx_retarr_tzoffset_hours = allocVector( REALSXP, nvals ));

			PROTECT( sx_single_retval = allocVector( VECSXP, 7 )); /* a list that has year, mo, day, etc, each of which are arrays */

			SET_VECTOR_ELT( sx_single_retval, 0, sx_retarr_year   );
			SET_VECTOR_ELT( sx_single_retval, 1, sx_retarr_month  );
			SET_VECTOR_ELT( sx_single_retval, 2, sx_retarr_day    );
			SET_VECTOR_ELT( sx_single_retval, 3, sx_retarr_hour   );
			SET_VECTOR_ELT( sx_single_retval, 4, sx_retarr_minute );
			SET_VECTOR_ELT( sx_single_retval, 5, sx_retarr_second );
			SET_VECTOR_ELT( sx_single_retval, 6, sx_retarr_tzoffset_hours );

			/* Elements are now part of a protected element, so they don't need protection */
			UNPROTECT(7);	/* sx_retarr_year, sx_retarr_month, sx_retarr_day, sx_retarr_hour, sx_retarr_minute, sx_retarr_second, sx_retarr_tzoffset_hours */

			/* Set our names */
			PROTECT( sx_name = allocVector( STRSXP, 7 ));
			SET_STRING_ELT( sx_name, 0, mkChar("year"  ) );
			SET_STRING_ELT( sx_name, 1, mkChar("month" ) );
			SET_STRING_ELT( sx_name, 2, mkChar("day"   ) );
			SET_STRING_ELT( sx_name, 3, mkChar("hour"  ) );
			SET_STRING_ELT( sx_name, 4, mkChar("minute") );
			SET_STRING_ELT( sx_name, 5, mkChar("second") );
			SET_STRING_ELT( sx_name, 6, mkChar("tzoffset_hours") );
			setAttrib( sx_single_retval, R_NamesSymbol, sx_name );
			UNPROTECT(1);	/* done with sx_name */
			}

		else
			error( "utCalendar_v3p0 (R version): error: passed unknown style, only 1 or 2 recognized!\n" );
		}

	for( i=0; i<nvals; i++ ) {
		if( (retval = utCalendar2_cal( *(value+i), unit_str, &year, &month,
				&day, &hour, &minute, &second, &f_tzoffset_hours, calendar )) != 0 ) {

			if( retval == UT_ENOINIT ) 
				error( "utCalendar (R version): error: udunits package not initialized yet!  You must call utInit() first." );

			else if( retval == UT_EINVALID )
				error( "utCalendar (R version): error: units are not time-like!" );

			else
				error( "utCalendar (R version): unknown error %d!\n", retval );
			}

		if( f_tzoffset_hours > 1.e20 ) 
			d_tzoffset_hours = NA_REAL;
		else
			d_tzoffset_hours = (double)f_tzoffset_hours;

		if( (nvals==1) || (*style==1) ) {
			/* Make this individual entry's utDate object */
			PROTECT( sx_single_retval = allocVector( VECSXP, 7 ));

			PROTECT( sx_year   = allocVector( INTSXP,  1 )); INTEGER(sx_year  )[0] = year;
			PROTECT( sx_month  = allocVector( INTSXP,  1 )); INTEGER(sx_month )[0] = month;
			PROTECT( sx_day    = allocVector( INTSXP,  1 )); INTEGER(sx_day   )[0] = day;
			PROTECT( sx_hour   = allocVector( INTSXP,  1 )); INTEGER(sx_hour  )[0] = hour;
			PROTECT( sx_minute = allocVector( INTSXP,  1 )); INTEGER(sx_minute)[0] = minute;
			PROTECT( sx_second = allocVector( REALSXP, 1 )); REAL   (sx_second)[0] = second;
			PROTECT( sx_tzoffset_hours = allocVector( REALSXP, 1 )); REAL   (sx_tzoffset_hours)[0] = d_tzoffset_hours;

			SET_VECTOR_ELT( sx_single_retval, 0, sx_year   );
			SET_VECTOR_ELT( sx_single_retval, 1, sx_month  );
			SET_VECTOR_ELT( sx_single_retval, 2, sx_day    );
			SET_VECTOR_ELT( sx_single_retval, 3, sx_hour   );
			SET_VECTOR_ELT( sx_single_retval, 4, sx_minute );
			SET_VECTOR_ELT( sx_single_retval, 5, sx_second );
			SET_VECTOR_ELT( sx_single_retval, 6, sx_tzoffset_hours );

			/* Elements are now part of a protected element, so they don't need protection */
			UNPROTECT(7);	/* sx_year, sx_month, sx_day, sx_hour, sx_minute, sx_second, sx_tzoffset_hours */

			/* Set our names */
			PROTECT( sx_name = allocVector( STRSXP, 7 ));
			SET_STRING_ELT( sx_name, 0, mkChar("year"  ) );
			SET_STRING_ELT( sx_name, 1, mkChar("month" ) );
			SET_STRING_ELT( sx_name, 2, mkChar("day"   ) );
			SET_STRING_ELT( sx_name, 3, mkChar("hour"  ) );
			SET_STRING_ELT( sx_name, 4, mkChar("minute") );
			SET_STRING_ELT( sx_name, 5, mkChar("second") );
			SET_STRING_ELT( sx_name, 6, mkChar("tzoffset_hours") );
			setAttrib( sx_single_retval, R_NamesSymbol, sx_name );
			UNPROTECT(1);	/* done with sx_name */

			/* Set our class name */
			PROTECT( sx_name = allocVector( STRSXP, 1 ));
			SET_STRING_ELT( sx_name, 0, mkChar("utDate"));
			setAttrib( sx_single_retval, R_ClassSymbol, sx_name );
			UNPROTECT(1);   /* done with sx_name */

			/* If only 1 value to convert, we are all done now! */
			if( nvals == 1 ) {
				UNPROTECT(1);
				return(sx_single_retval);
				}

			/* ...otherwise, add this new utDate to the list we are accumulating */
			SET_VECTOR_ELT( sx_big_retval, i, sx_single_retval );
			UNPROTECT(1);	/* sx_single_retval is now protected by its parent */
			}
		else
			{
			/* style == 2 */
			INTEGER(sx_retarr_year  )[i] = year;
			INTEGER(sx_retarr_month )[i] = month;
			INTEGER(sx_retarr_day   )[i] = day;
			INTEGER(sx_retarr_hour  )[i] = hour;
			INTEGER(sx_retarr_minute)[i] = minute;
			REAL   (sx_retarr_second)[i] = second;
			REAL   (sx_retarr_tzoffset_hours)[i] = d_tzoffset_hours;
			}
		}

	/* NOTE! if nvals==1, we exit ABOVE, not here */
	if( *style == 1 ) {
		UNPROTECT(1);
		return( sx_big_retval );
		}
	else
		{
		UNPROTECT(1);
		return( sx_single_retval );
		}
}

/**********************************************************************************************************/
/* Converts a given year, month, day, hour, minute, and second
 * into a value, given an appropriate time unit to work with.
 * Inputs:
 *	ndates: number of dates to convert
 *	year: array of years
 *	month: array of months
 *	day: array of days
 *	hour: array of hours
 *	minute: array of minutes
 *	second: array of seconds (double precision value)
 * Outputs:
 *	a list with elements $error and $values.
 *	$values correspond to the passed date(s) in the given units
 *	$error is 0 if no error, and not zero if an error occurred.
 */
SEXP R_utInvCalendar_ccs( SEXP sx_ndates, SEXP sx_year, SEXP sx_month, SEXP sx_day, SEXP sx_hour,
	SEXP sx_minute, SEXP sx_second, SEXP sx_unit_str, SEXP sx_calendar ) 
{
	ut_unit	*u;
	int 	i, ierr, nvals;
	SEXP	sx_error, sx_values, sx_retval, sx_retnames;
	int	*year, *month, *day, *hour, *minute;
	double	*second, tval;
	const char *calendar = CHAR(STRING_ELT(sx_calendar,0));
	const char *unit_str = CHAR(STRING_ELT(sx_unit_str,0));

	/* Make our returned list, which has elements $values and $error */
	PROTECT( sx_retval = allocVector( VECSXP, 2 ));

	/* Set the names for the returned list */
	PROTECT( sx_retnames = allocVector( STRSXP, 2 ));       /* 2 elements in the returned list */
	SET_STRING_ELT( sx_retnames, 0, mkChar("error"));
	SET_STRING_ELT( sx_retnames, 1, mkChar("values" ));
	setAttrib( sx_retval, R_NamesSymbol, sx_retnames );
	UNPROTECT(1);						/* done with sx_retnames */

	/* Set the return error code */
	PROTECT( sx_error = allocVector( INTSXP, 1 ));
	INTEGER( sx_error)[0] = -1;				/* Set error return code to -1 (error) to begin */
	SET_VECTOR_ELT( sx_retval, 0, sx_error   );

	nvals = length(sx_year);	/* The way this is called, guaranteed length of all input sxp's is equal */

	year   = INTEGER(sx_year  );
	month  = INTEGER(sx_month );
	day    = INTEGER(sx_day   );
	hour   = INTEGER(sx_hour  );
	minute = INTEGER(sx_minute);
	second = REAL   (sx_second);

	/* Make our returned value array */
	PROTECT( sx_values = allocVector( REALSXP, nvals ));

	for( i=0; i<nvals; i++ ) {
		if( (ierr = utInvCalendar2_cal( year[i], month[i], day[i],
				hour[i], minute[i], second[i], unit_str, &tval, calendar )) != 0 ) {
			if( ierr == UT_ENOINIT ) {
				Rprintf( "utInvCalendar (R version): error: udunits package not initialized yet!\n");
				Rprintf( "You must call utInit() first.\n" );
				UNPROTECT(3);
				return( sx_retval );
				}

			if( ierr == UT_EINVALID ) {
				Rprintf( "utInvCalendar (R version): error: units are not temporal!\n" );
				UNPROTECT(3);
				return( sx_retval );
				}

			Rprintf( "utInvCalendar (R version): unknown error %d!\n", ierr );
			UNPROTECT(3);
			return( sx_retval );
			}
		REAL(sx_values)[i] = tval;
		}

	/* Finish assembling output list */
	SET_VECTOR_ELT( sx_retval, 1, sx_values );
	UNPROTECT(3);
	INTEGER(sx_error)[0] = 0;			/* "No error" return */
	return( sx_retval );
}


