#=========================================================================================
# Returns the file size in bytes, or -1 if the file was not found
#
filesize <- function( fname, verbose=FALSE ) {

        #if( dir_exists( fname )) stop(paste("error, was called with a DIRECTORY name, not a FILE name:", fname ))

        if( length( fname ) > 1 ) stop(paste("error, can only handle 1 filename at the moment, here was input list:", pv(fname) ))

        if( verbose ) print(paste("filesize: entering with", fname))

        cmd <- paste('ls -l', fname)
        if( verbose ) print(paste("filesize: cmd:", cmd ))
        q   <- system( cmd, intern=TRUE, ignore.stderr=TRUE )
        if( verbose ) print(paste("filesize: length(q) (return value):", length(q) ))
        if( verbose ) print(paste("filesize: q (return value):", q ))
        if( length(q) == 0 ) {
                if( verbose ) print(paste("filesize: returning -1"))
                return(-1)
                }

        if( verbose ) print(paste("filesize: strsplit"))
        t <- strsplit( q, ' ' )
        if( verbose ) {
                print(paste("filesize: t:"))
                print(t)
                }
        csz = t[[1]][5]
        ncsz = nchar( csz )
        if( ncsz > 9 )
                sz = as.double( csz )
        else
                sz = as.integer(csz)

        return( sz )
}

#===============================================================================
# If the file looks like this:
#
# 1900  12.3
# 1901  17.9
# 1902  8.7
# 1903  17.5
# ...
# 2011  15.4
#
# then the shape of the returned array will be [nt,2]
# I.e., it is [nlines,ncols]
#
reads <- function( fname, skip=0, numbers_have_D_not_e=FALSE, is_char=FALSE, nrows=-1 ) {

        if( filesize( fname ) < 1 )
                stop(paste("Error, reads called on non-existant file:", fname ))

        t <- read.table(fname, skip=skip, stringsAsFactors=FALSE, nrows=nrows )
        ncols <- length(t)
        nrows <- length(t[[1]])

        if( ncols == 1 ) {
                if( numbers_have_D_not_e ) {
                        z = data.matrix(t$V1)
                        z = gsub( 'D', 'e', z )
                        x = as.double(z)
                        }
                else
                        {
                        if( is_char )
                                x <- t$V1
                        else
                                x <- as.double(t[[1]])
                        }
                }
        else
                {
                if( is_char ) {
                        x <- array(' ',dim=c(nrows,ncols))
                        for( j in 1:ncols )
                                x[,j] <- t[[j]]
                        }
                else
                        {
                        x <- array(0.,dim=c(nrows,ncols))
                        for( j in 1:ncols )
                                x[,j] <- as.double( t[[j]] )
                        }
                }

        return(x)
}

#===========================================================================================
# Returns a ncvar structure (list)
#
nc_main_var = function( ncid, debug=FALSE ) {

        if( debug ) print(paste('nc_main_var: entering for file', ncid$filename ))

        nv = ncid$nvars

        if( nv == 1 ) {
                if( debug ) print(paste('nc_main_var: just one var in file, returning:', ncid$var[[1]]$name ))
                return( ncid$var[[1]] )
                }

        vsize = array(0., nv)
        for( ivar in 1:nv ) {
                nd = ncid$var[[ivar]]$ndims
                if( nd > 0 ) {
                        size = 1
                        for( idim in 1:nd )
                                size = size * ncid$var[[ivar]]$dim[[idim]]$len
                        vsize[ivar] = size
                        }
                else
                        vsize[ivar] = 1
                if( debug ) print(paste("nc_main_var: var", ivar, '/', nv, "=", ncid$var[[ivar]]$name, 'size:', vsize[ivar] ))
                }

        im = which.max(vsize)

        if( debug ) print(paste('nc_main_var: max size is index', im, "size", vsize[im], 'var', ncid$var[[im]]$name ))
        return( ncid$var[[ im ]] )
}

#=========================================================================================
# This makes a blank 'canvas' to later draw points and lines on.
#
canvas = function( xlim=NA, ylim=NA, type='b', pch='.', bg=NA, box=TRUE, xlab='', is_geographical=FALSE, ... ) {

        globalframenumber <<- globalframenumber + 1

        if( is.na(xlim) && is.na(ylim)) {
                #---------------------------------------------------
                # A TRULY blank canvas, for skipping a plotting spot
                #---------------------------------------------------
                mplot( c(100,101), c(100,101), xlim=c(0,1), ylim=c(0,1), pch='.', col='white', axes=FALSE, box=FALSE, grid=FALSE, xlab='' )
                return()
                }

        mplot( c(-10000.1,-10000), c(-10000.1,-10000), type=type, xlim=xlim, ylim=ylim, box=box, xlab=xlab, ... )

        if( ! is.na(bg)) {
                usr = par()$usr   # vector of the extrems of the user coordinates of the PLOTTING region
                x0 = usr[1]
                x1 = usr[2]
                y0 = usr[3]
                y1 = usr[4]
                polygon( c(x0,x1,x1,x0), c(y0,y0,y1,y1), col=bg, border=bg )
                }

        #---------------------------------------------------------
        # Check the geometry of the plot and make sure it's square
        # We assume X and Y are in degrees
        # Note: aspect ratio is width/height
        #---------------------------------------------------------
        if( is_geographical ) check_plot_aspect_ratio( 'canvas', xlim, ylim )
}

#=============================================================================
mplot <- function(x,y,...,grid=TRUE,xvals=NA,
			gglook=FALSE,
			posnegbars=NA, barbase=0., posnegbar_pos='red', posnegbar_neg='blue',
			xlab='',
			ylab='',
			lty=1,
			xaxs='r',
			yaxs='r',
			xlabline=1.3,
			ylabline=2,
			cex_xlabline=0.9,
			cex_ylabline=0.9,
			type='l',
			col.axis='red',
			lty.grid=3,
			subgridx=NA,
			subgridy=NA,
			col.grid=rgb(0.5,0.5,0.5),
			axes=NA,
			box=TRUE,
			yxform_func=NA,	  # a function to transform the Y cordinate for plotting purposes
			xlim=NA, ylim=NA,
			gmtborder=NA,  # set to fractional width of border (try 0.01)
			saveplot=FALSE,
			mgp_x_axis=0.,
			loadplot='',
			incr_global_frameno=TRUE,
			is_geographical=FALSE,
			yvals=NA,ylabels=NA) {

	trace <- FALSE

	if( trace ) {
		print("mplot: entering" )
		print(paste("mplot: ylim=", pv(ylim) ))
		}

	if( sum(is.na(y)) == length(y)) {
		print(paste("mplot: error, called with all NA Y values, n=", length(y) ))
		canvas( xlim=range(x, na.rm=TRUE), ylim=c(0,1), ... )
		s = paste( 'All Y values are NA (n=', length(y), ')' )
		text( min(x, na.rm=TRUE), 0.5, s, adj=c(0.0, 0.5))
		return()
		}

	if( length( ylim ) == 2 ) {
		if( is.na(ylim[1]) || (is.na(ylim[2]))) 
			stop(paste("Error, called mplot with bad ylim:", pv(ylim) ))
		if( (! is.finite(ylim[1])) || (! is.finite(ylim[2]))) 
			stop(paste("Error, called mplot with bad (not finite) ylim:", pv(ylim) ))
		}

	if( ! exists('globalframenumber'))
		globalframenumber <<- 0

	if( incr_global_frameno ) 
		globalframenumber <<- globalframenumber + 1

	if( class(x) == 'SpatialPolygons' ) {
		np1 = length( x@polygons )
		print(paste('npoly level 1:', np1))
		for( ipoly in 1:np1 ) {
			np2 = length( x@polygons[[ipoly]]@Polygons )
			print(paste('For polygong', ipoly, ' # lev 2 polygons:', np2 ))
			for( ip2 in 1:np2 ) {
				txx = x@polygons[[ipoly]]@Polygons[[ip2]]@coords[,1]
				tyy = x@polygons[[ipoly]]@Polygons[[ip2]]@coords[,2]
				print(paste("Calling mplot with iouter, inner polygons:", ipoly, ip2 ))
				mplot( txx, tyy, main='SP mplot', ... )
				}
			}

		return()
		}

	if( saveplot ) {
		saveplot_fname = paste( myname, ".frame_", mformatC('%02d',globalframenumber), '.Rdat', sep='' )
		sav   		= list()
		sav$x 		= x
		sav$y 		= y
		sav$grid 	= grid
		sav$xlab 	= xlab
		sav$ylab 	= ylab
		sav$ylabline	= ylabline
		sav$type	= type
		sav$col		= col
		sav$lty		= lty
		sav$axes	= axes
		sav$box		= box
		sav$xlim	= xlim
		sav$ylim	= ylim
		sav$xaxs	= xaxs
		sav$yaxs	= yaxs
		save( file=saveplot_fname, sav )
		}
	else if( nchar(loadplot) > 0 ) {
		load( loadplot )
		x 		= sav$x
		y 		= sav$y
		grid 		= sav$grid
		xlab 		= sav$xlab
		ylab 		= sav$ylab
		ylabline	= sav$ylabline
		type		= sav$type
		col		= sav$col
		lty		= sav$lty
		axes		= sav$axes
		box		= sav$box
		xlim		= sav$xlim
		ylim		= sav$ylim
		xaxs		= sav$xaxs
		yaxs		= sav$yaxs
		}

	if( (length(y) == 1) && (is.na(y)) )
		stop(paste('mplot: error, must supply y values!'))

	if( length(x) != length(y)) {
		print("Error in mplot, have different number of X and Y values!")
		print(paste("# X values:",length(x)))
		print(paste("# Y values:",length(y)))
		stop('X and Y must be same length')
		}

	if( trace )
		print("mplot: ==a" )

	if( (length(xlim)==1) && (is.na(xlim)) )
		xlim <- range(x, na.rm=TRUE )
	if( (length(ylim)==1) && (is.na(ylim)) )
		ylim <- range(y, na.rm=TRUE )

	if( is.na(ylim[1]) || is.na(ylim[2])) {
		print(paste("Error, called mplot with NA in the ylim! This is not allowed"))
		print(paste("ylim:", pv(ylim) ))
		quickstats( x, 'xvals' )
		quickstats( y, 'yvals' )
		stop('exiting now')
		}

	#---------------------------------------------------------
	# Check the geometry of the plot and make sure it's square
	# We assume X and Y are in degrees
	# Note: aspect ratio is width/height
	#---------------------------------------------------------
	if( is_geographical ) check_plot_aspect_ratio( 'ccontour', xlim, ylim )

	if( trace )
		print("mplot: ==b" )

	if( !is.na(yxform_func)) {
		if( is.na(yvals) ) 
			stop("if specify yxform_func, must specify yvals as well (values to label)")
		y <- yxform_func(y)
		axes <- FALSE
		ylim <- yxform_func(ylim)
		}

	if( trace )
		print("mplot: ==c" )

	par(tck=0.02,las=1,xaxs=xaxs,yaxs=yaxs)

	if( ! is.na(axes)) 
		myaxes <- axes
	else
		myaxes <- TRUE
	if( (!is.na(xvals)) || (!is.na(yvals)) ) {
		myaxes <- FALSE
		}

	if( trace )
		print("mplot: ==d" )

	if( ! is.na(posnegbars) ) {
		if( trace )
			print("mplot: ==d2" )
		nx <- length(x)
		if( length(posnegbars) == nx )
			barcolors = posnegbars
		else
			barcolors = ifelse( y >= 0, posnegbar_pos, posnegbar_neg )
		xtics <- pretty(x)
		ytics <- pretty(y)
		mnx <- min(x)
		mxx <- max(x)
		mny <- min(y)
		mxy <- max(y)
		plot(c(mnx,mxx),c(mny,mxy),type='p',pch='.',col='white',...,xlim=xlim,ylim=ylim)
		dx <- x[2]-x[1]
		for( i in 1:nx ) {
			xleft  <- x[i]-dx/2.
			xright <- xleft + dx
			ybot   <- barbase
			ytop   <- y[i]
			polygon(c(xleft,xright,xright,xleft),
				c(ybot,ybot,ytop,ytop), 
				border=barcolors[i], col=barcolors[i] )
			}
		}
	else
		{
		if( trace )
			print("mplot: ==d5" )

		#--------------------------------------------------------
		# HERE IS WHERE WE MAKE THE MAIN CANVAS AND PLOT THE LINE
		#--------------------------------------------------------
		plot(x,y,xlab='',ylab='', lty=lty, ...,axes=FALSE,type=type,xlim=xlim,ylim=ylim,bty='n')

		if( gglook ) {
			uu = par()$usr	# llx, urx, lly, ury
			llx = uu[1]
			urx = uu[2]
			lly = uu[3]
			ury = uu[4]
			polygon( c(llx,llx,urx,urx), c(lly,ury,ury,lly), col=grey(0.9), border=NA )

			xaxp = par()$xaxp
			print(paste('xaxp:', pv(xaxp) ))

			#-------------------
			# White X grid lines
			#-------------------
			txrange = xaxp[2] - xaxp[1]
			tdx = txrange / xaxp[3]
			txtics = seq( xaxp[1], xaxp[2], by=tdx )
			tnxt = length( txtics )
			for( ii in 1:tnxt ) 
				abline( v=txtics[ii], col='white', lwd=2 )

			# Do half tics in X
			for( ii in 1:tnxt ) 
				abline( v=txtics[ii]+tdx/2, col=grey(0.95), lwd=1 )

			yaxp = par()$yaxp
			print(paste('yaxp:', pv(yaxp) ))

			#-------------------
			# White Y grid lines
			#-------------------
			tyrange = yaxp[2] - yaxp[1]
			tdy = tyrange / yaxp[3]
			tytics = seq( yaxp[1], yaxp[2], by=tdy )
			tnyt = length( tytics )
			for( ii in 1:tnyt ) 
				abline( h=tytics[ii], col='white', lwd=2 )

			# Do half tics in X
			for( ii in 1:tnyt ) 
				abline( h=tytics[ii]+tdy/2, col=grey(0.95), lwd=1 )

			lines( x, y, lty=lty, ..., type=type )
			}

		if( myaxes ) {
			#---------------------------------------------------------------------
			# Adjust the offset of the X axis TIC MARKS (not label) to be less
			# than the default of mgp[2] = 0.2. This is awkward because that value
			# works fine for the Y axis, but is too large for the X axis. So, have
			# do do the axes separately
			#---------------------------------------------------------------------
			mgp_orig = par()$mgp
			mgp = mgp_orig
			mgp[2] = mgp_x_axis
			par(mgp=mgp)
			axis( 1 )

			par(mgp=mgp_orig)
			par(las=1)
			axis(2)
			}
		}

	#---------------------------------------------
	# Put on the X axes LABEL (not the tick marks)
	#---------------------------------------------
	mtext(xlab,1,line=xlabline,cex=cex_xlabline)

	if( trace )
		print("mplot: ==e" )

	par(las=0)
	mtext(ylab,2,line=ylabline,cex=cex_ylabline)

	if( trace )
		print("mplot: ==f" )

	if( grid ) {
		if( trace )
			print("mplot: ==g" )

		log <- par("xlog")
		ax <- par("xaxp")
		at <- axTicks(1, axp = ax, log=log)
		nxticks <- length(at)
		for( i in 1:nxticks )
			abline( v=at[i], lty=lty.grid, col=col.grid )
		if( !is.na(subgridx)) {
			delat <- at[2]-at[1]
			at <- c(at[1]-delat,at,at[nxticks]+delat)
			nxticks <- length(at)
			for( i in 1:(nxticks-1) ) {
				dx <- (at[i+1]-at[i])/subgridx
				for( st in 1:(subgridx-1) ) 
					abline(v=at[i]+dx*st, lty=lty.grid, col=col.grid )
				}
			}

		if( is.na(yxform_func) ) {
			log <- par("ylog")
			ax <- par("yaxp")
			at <- axTicks(2, axp = ax, log=log)
			nyticks <- length(at)
			for( i in 1:nyticks )
				abline( h=at[i], lty=lty.grid, col=col.grid )
			if( !is.na(subgridy)) {
				delat <- at[2]-at[1]
				at <- c(at[1]-delat,at,at[nyticks]+delat)
				nyticks <- length(at)
				for( i in 1:(nyticks-1) ) {
					dy <- (at[i+1]-at[i])/subgridy
					for( st in 1:(subgridy-1) ) {
						abline(h=at[i]+dy*st, lty=lty.grid, col=col.grid )
						}
					}
				}
			}

		}
	else
		{
		if( trace )
			print("mplot: ==h" )

		log <- par("ylog")
		ax <- par("yaxp")
		at <- axTicks(2, axp = ax, log=log)
		nyticks <- length(at)
		if( !is.na(subgridy)) {
			delat <- at[2]-at[1]
			at <- c(at[1]-delat,at,at[nyticks]+delat)
			nyticks <- length(at)
			dxtic <- (xlim[2]-xlim[1])*.02
			for( i in 1:(nyticks-1) ) {
				dy <- (at[i+1]-at[i])/subgridy
				for( st in 1:(subgridy-1) ) {
					abline(h=at[i]+dy*st, lty=lty.grid, col=col.grid )
					}
				}
			}
		}

	if( trace )
		print("mplot: ==i" )

	if( ! is.na(xvals) ) {
		if( is.na(xlabels) ) 
			labelToUse <- T
		else
			labelToUse <- xlabels
		axis(1,xvals,labels=labelToUse)
		axis(3,xvals,labels=F)

		if( is.na(yvals) ) {
			axis(2)
			axis(4,labels=F)
			}

		if( grid )
			abline(v=xvals,lty=3)
		}

	if( trace )
		print("mplot: ==j" )

	if( ! is.na(yvals) ) {
		if( is.na(yxform_func)) {
			if( is.na(ylabels) ) 
				labelToUse <- T
			else
				labelToUse <- ylabels
			axis(2,yvals,labels=labelToUse,col.axis=col.axis)

			if( is.na(xvals) ) {
				axis(1)
				axis(3,labels=F)
				}
			if( grid )
				abline(h=yvals,lty=3)
			}
		else
			{
			ny <- length(yvals)
			width <- xlim[2]-xlim[1]
			xoffset <- xlim[1] - width*.04
			for( ii in 1:ny ) {
				yoffset <- yxform_func( yvals[ii] )
				text(xoffset,yoffset,yvals[ii],xpd=NA, adj=c(1.2,0.5))
				if( grid )
					abline(h=yoffset,lty=lty.grid, col=col.grid )
				# tic
				lines(c(xoffset,xoffset+width*.01), c(yoffset,yoffset))
				}
			}
		}

	if( trace )
		print("mplot: ==k" )

	if( !is.na(yxform_func))
		axis(1)

	if( box )
		box( col=par()$col.axis )

	if( trace )
		print("mplot: ==l : exiting" )

	if( !is.na(gmtborder) ) {
		addgmtborder( gmtborder )
		}
}

#=============================================================================
# Returns the indes of the value in matrix 'matrix' that is closest to the
# passed target value 'value'.  If a minimum distance is specified, and no
# value is that close, -1 is returned.  If there are multiple matches, the
# index of the first match is returned.
#
wherenearest <- function( value, matrix, mindist=NA, verbose=FALSE ) {

        nn <- length(matrix)
        if( verbose )
                print(paste("------------ entering wherenearest: looking for", value, " in array from ", matrix[1], "to",
                                matrix[nn]))

        nmiss = sum( is.na(matrix))
        if( nmiss == nn ) stop(paste("error: supplied vector of values is all NA's! Supplied single value was:", value ))

        dist  <- abs(matrix-value)
        wmidx <- which.min( dist )      # returns index of FIRST minimum
        if( verbose ) print(paste( "wherenearest: closest value is at index=", wmidx,
                        "which have value", matrix[wmidx]))

        if( is.na(mindist))
                return( wmidx )

        minval <- dist[wmidx]
        if( verbose )
                print(paste("wherenearest: using mindist=", mindist,
                        "and distance at min index is", minval))

        if( minval > mindist ) {
                if( verbose ) print('wherenearest: mnidist criterion fails, returning -1')
                return(-1)
                }
        else
                return( wmidx )
}


